/* for integer types */
#include <stdint.h>

static uint8_t lcram[16384];
static int lcread, lcwrite, lcalt;

/*
 * Module information
 */
__declspec(dllexport) uint32_t mod_type=0x544F4C53;
__declspec(dllexport) uint8_t *mod_name="16K Language Card";
__declspec(dllexport) uint8_t *mod_pref="lcard";
__declspec(dllexport) uint8_t *mod_vers="0.01";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2017 Steve Nickolas\n";

/*
 * Tell the shell to load our ROM for us
 */
__declspec(dllexport) uint8_t  *lcard_preload_name="";
__declspec(dllexport) uint16_t lcard_preload_size=0;
__declspec(dllexport) uint8_t  *lcard_preload_ptr=0;

static void lcard_do (uint16_t a)
{
 switch (a&0x0F)
 {
  case 0x0:
  case 0x4:
   lcalt=0x80;
   lcread=0x80;
   lcwrite=0x00;
   break;
  case 0x1:
  case 0x5:
   lcalt=0x80;
   lcread=0x00;
   lcwrite=0x80;
   break;
  case 0x2:
  case 0x6:
   lcalt=0x80;
   lcread=0x00;
   lcwrite=0x00;
   break;
  case 0x3:
  case 0x7:
   lcalt=0x80;
   lcread=0x80;
   lcwrite=0x80;
   break;
   
  case 0x8:
  case 0xC:
   lcalt=0x00;
   lcread=0x80;
   lcwrite=0x00;
   break;
  case 0x9:
  case 0xD:
   lcalt=0x00;
   lcread=0x00;
   lcwrite=0x80;
   break;
  case 0xA:
  case 0xE:
   lcalt=0x00;
   lcread=0x00;
   lcwrite=0x00;
   break;
  case 0xB:
  case 0xF:
   lcalt=0x00;
   lcread=0x80;
   lcwrite=0x80;
   break;
 }
}

__declspec(dllexport) void lcard_write16 (uint16_t a, uint8_t v)
{
 lcard_do (a); 
}

/*
 * We are a slot 0 card; we don't have a rom space.
 */
__declspec(dllexport) void lcard_write256 (uint16_t a, uint8_t v)
{
}

/*
 * We don't have a shared memory space.
 */
__declspec(dllexport) void lcard_write2048 (uint16_t a, uint8_t v)
{
}

__declspec(dllexport) uint8_t lcard_read16 (uint16_t a)
{
 lcard_do (a); 
 return 255;
}

/*
 * We are a slot 0 card; we don't have a rom space.
 */
__declspec(dllexport) uint8_t lcard_read256 (uint16_t a)
{
 return 255;
}

/*
 * We don't have a shared memory space.
 */
__declspec(dllexport) uint8_t lcard_read2048 (uint16_t a)
{
 return 255;
}

/*
 * No change to what writing to any memory address will do.
 */
__declspec(dllexport) int16_t lcard_override_wr (uint16_t a, uint8_t v)
{
 if (a<0xD000) return -1;
 if (!lcwrite) return -1;
 if (a<0xE000)
 {
  if (lcalt)
  {
   lcram[a-0xD000]=v;
   return 0;
  }
 }
 lcram[a-0xC000]=v;
 return 0;
}

/*
 * Intercept READS to the top 12K of memory, IF enabled.
 * Otherwise do nothing.
 */
__declspec(dllexport) int16_t lcard_override_rd (uint16_t a)
{
 if (a<0xD000) return -1;
 if (!lcread) return -1;
 if (a<0xE000)
 {
  if (lcalt) return lcram[a-0xD000];
 }
 return lcram[a-0xC000];
}

/*
 * Disable ourself when Reset is hit.
 */
__declspec(dllexport) int lcard_reset (void)
{
 lcread=lcwrite=lcalt=0;
 return 0;
}

/*
 * Start out disabled.
 */
__declspec(dllexport) int lcard_init (void)
{
 lcread=lcwrite=lcalt=0;
 return 0;
}

/*
 * Nothing to do when deinitializing.
 */
__declspec(dllexport) int lcard_deinit (void)
{
 return 0;
}

__declspec(dllexport) void lcard_config (void)
{
}
