#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

static char *verstr=
  "MoDapple  Version 0.10\n"
  "Copyright 2016 Steve Nickolas\n"
  "Other copyrights may apply to plugin code; see below for details.\n";

typedef struct
{
 HMODULE  dll;
 uint32_t *mod_type;
 uint8_t  **mod_name;
 uint8_t  **mod_pref;
 uint8_t  **mod_vers;
 uint8_t  **mod_copr;
 uint8_t  **preload_name;
 uint16_t *preload_size;
 uint8_t  **preload_ptr;
 void     (*write16)     (uint16_t a, uint8_t v);
 void     (*write256)    (uint16_t a, uint8_t v);
 void     (*write2048)   (uint16_t a, uint8_t v);
 uint8_t  (*read16)      (uint16_t a);
 uint8_t  (*read256)     (uint16_t a);
 uint8_t  (*read2048)    (uint16_t a);
 int16_t  (*override_rd) (uint16_t a);
 int16_t  (*override_wr) (uint16_t a, uint8_t v);
 int      (*reset)       (void);
 
 /*
  * Formerly, init() was a boolean.
  * For plugins intended for MoDapple 0.09 and later, if init() returns "6502"
  * then it needs pointers to cpuread() and cpuwrite().  The function to take
  * these pointers is only linked if it is required.
  */
 int      (*init)        (void);
 int      (*deinit)      (void);
 
 void     (*dmainit)     (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t));

 void     (*config)      (void);
} SLOT;

SLOT slot[8];

HINSTANCE hinstance;

/* Not currently used, but will affect i/o reads and writes */
uint8_t  iie_128K, iie_store80, iie_rdaux, iie_wraux, iie_lcaux;

uint8_t  ram[49152], rom[12288];

uint8_t  fnstr[256];

/* Data pointers and functinos for TTY driver */
uint32_t *tty_mod_type;
uint8_t  **tty_mod_name;
uint8_t  **tty_mod_pref;
uint8_t  **tty_mod_vers;
uint8_t  **tty_mod_copr;
uint8_t  *ttykey;
void     (*ttywrite)   (uint16_t a, uint8_t v);
uint8_t  (*ttyread)    (uint16_t a);
void     (*ttyout)     (uint8_t a, uint8_t v);
uint8_t  (*ttyin)      (uint8_t a);
void     (*ttydeinit)  (void);
void     (*ttyinit)    (void);
int      (*ttyloop)    (void);

/* Data pointers and functinos for CPU driver */
uint32_t *cpu_mod_type;
uint8_t  **cpu_mod_name;
uint8_t  **cpu_mod_pref;
uint8_t  **cpu_mod_vers;
uint8_t  **cpu_mod_copr;
void     (*cpureset)   (void);
void     (*cpurun)     (void);
void     (*cpuinit)    (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t), int8_t (*v_cpuloop)(void));
void     (*cpuseta)    (uint8_t val);
void     (*cpusetx)    (uint8_t val);
void     (*cpusety)    (uint8_t val);
void     (*cpusets)    (uint8_t val);
void     (*cpusetp)    (uint8_t val);
void     (*cpusetpc)   (uint16_t val);
void     (*cpusetrate) (int val);
uint8_t  (*cpugeta)    (void);
uint8_t  (*cpugetx)    (void);
uint8_t  (*cpugety)    (void);
uint8_t  (*cpugets)    (void);
uint8_t  (*cpugetp)    (void);
uint16_t (*cpugetpc)   (void);
int      (*cpugetrate) (void);

HMODULE mod_tty, mod_cpu;

void notamod (char *header, char *filename)
{
 if (mod_tty) FreeLibrary(mod_tty);
 if (mod_cpu) FreeLibrary(mod_cpu);
 fprintf (stderr, "%s: %s is not a MoDapple module.\n", header, filename);
 exit(1);
}

void nofunc (char *header, char *filename, char *ident)
{
 if (mod_tty) FreeLibrary(mod_tty);
 if (mod_cpu) FreeLibrary(mod_cpu);
 fprintf (stderr, "%s: Could not find a '%s' function in %s.\n", header, ident, filename);
 exit(1);
}

int nofunc2 (char *header, int slotno, char *filename, char *ident)
{
 FreeLibrary(slot[slotno].dll);
 slot[slotno].dll=0;
 fprintf (stderr, "%s: Could not find a '%s' function in %s.\n", header, ident, filename);
 return -1;
}

#define TFUNCTINO(a,b) sprintf (fnstr,"%s_%s",*tty_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_tty,fnstr))) nofunc(progname, ttymod, b)
#define CFUNCTINO(a,b) sprintf (fnstr,"%s_%s",*cpu_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_cpu,fnstr))) nofunc(progname, cpumod, b)
#define SFUNCTINO(s,a,b) sprintf (fnstr,"%s_%s",*(slot[s].mod_pref),b); if (!(slot[s].a=(void*)GetProcAddress(slot[s].dll,fnstr))) return nofunc2(progname, s, module, b)

char *progname="MoDapple";

int bload (char *filename, int expect, uint8_t *pointer)
{
 FILE *file;
 long l;
 
 file=fopen(filename,"rb");
 if (!file)
 {
  fprintf (stderr, "%s: ", progname);
  perror(filename);
  return -1;
 }
 
 fseek(file,0,SEEK_END);
 l=ftell(file);
 fseek(file,0,SEEK_SET);
 if (l!=expect)
 {
  fclose(file);
  fprintf (stderr, "%s: %s: Wrong file size (%lu expected, %lu found)\n", progname, filename, expect, l);
  return -1;
 }
 l=fread(pointer,1,expect,file);
 fclose(file);
 if (l<expect)
 {
  fprintf(stderr, "%s: %s: Short read (%lu expected, %lu found)\n", progname, filename, expect, l);
  return -1;
 }
 
 printf ("%s: Opened binary blob %s (%lu bytes)\n", progname, filename, expect);
 return 0;
}

static int holdc8;

void cpuwrite(uint16_t a, uint8_t v)
{
 int t, o;
 for (t=7; t>=0; t--)
 {
  if (slot[t].dll)
  {
   o=slot[t].override_wr(a, v);
   if (!o) return;
  }
 }

 if ((a>1023)&&(a<3072))  {ttywrite(a,v);}
 if ((a>8191)&&(a<24576)) {ttywrite(a,v);}
 if (a<49152)             {ram[a]=v; return;}
 
 /* XXX: does not currently implement the 2K shared memory at C800-CFFF */
 if ((a&0xFF80)==0xC080)
 {
  int s;
  s=a&0x0070;
  s>>=4;
  if (slot[s].dll)
  {
   slot[s].write16(a,v);
   return;
  }
 }
 if (((a&0xF800)==0xC000)&&(a>=0xC100))
 {
  int s;
  s=a&0x0700;
  s>>=8;
  holdc8=s;
  if (slot[s].dll)
  {
   slot[s].write256(a,v);
   return;
  }
 }

 if (((a&0xF800)==0xC800)&&holdc8)
 {
  if (!slot[holdc8].dll) return;
  slot[holdc8].write2048(a-0xC800,v);
  if (a==0xCFFF) holdc8=0;
  return;
 }

 if (a>=0xD000)
 {
  return;
 }
 
 if ((a&0xFFF0)==0xC000)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC010)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC030)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC050)  {ttyout(a,v); return;}
}

uint8_t cpuread(uint16_t a)
{
 int t, o;
 for (t=7; t>=0; t--)
 {
  if (slot[t].dll)
  {
   o=slot[t].override_rd(a);
   if (o>=0) 
   {
    return o;
   }
  }
 }

 if ((a>1023)&&(a<3072))  return ttyread(a);
 if ((a>8191)&&(a<24576)) return ttyread(a);
 if (a<49152)             return ram[a];
 
 if ((a&0xFF80)==0xC080)
 {
  int s;
  s=a&0x0070;
  s>>=4;
  if (slot[s].dll)
  {
   return slot[s].read16(a);
  }
  else return 0;
 }
 if (((a&0xF800)==0xC000)&&(a>=0xC100))
 {
  int s;
  s=a&0x0700;
  s>>=8;
  holdc8=s;
  if (slot[s].dll)
  {
   return slot[s].read256(a);
  }
  return rand()%256; /* well, it does the job */
 }

 if (((a&0xF800)==0xC800)&&holdc8)
 {
  uint8_t c;
  
  if (!slot[holdc8].dll) return 255;
  c=slot[holdc8].read2048(a-0xC800);
  if (a==0xCFFF) holdc8=0;
  return c;
 }

 if (a>=0xD000)
 {
  return rom[a-0xD000];
 }
 
 if ((a&0xFFF0)==0xC000)  return ttyin(a);
 if ((a&0xFFF0)==0xC010)  return ttyin(a);
 if ((a&0xFFF0)==0xC030)  return ttyin(a);
 if ((a&0xFFF0)==0xC050)  return ttyin(a);
 
 return 0;
}

static int die;

void reset_all(void)
{
 int t;
 
 cpureset();
 for (t=7; t>=0; t--)
 {
  if (slot[t].dll)
   slot[t].reset();
 }
}

int8_t cpuloop(void)
{
 int x;
 
 x=ttyloop();
 if (x==-1) return -1;
 if (x==1) reset_all();
 if ((x>=10)&&(x<18))
 {
  if (slot[x-10].dll)
  {
   slot[x-10].config();
  }
 }
 return 0;
}

int slotassign (int slotno, char *module)
{
 int e;
 
 if ((slotno<0)||(slotno>7))
 {
  fprintf (stderr, "%s: Invalid slot %d\n", progname, slotno);
  return -1;
 }
 
 slot[slotno].dll=LoadLibrary(module);
 if (!(slot[slotno].dll))
 {
  fprintf(stderr, "%s: Could not open module %s\n", progname, module);
  return -1;
 }

 printf ("%s: Opened module %s\n", progname, module);
 if (!(slot[slotno].mod_type=(void*)GetProcAddress(slot[slotno].dll,"mod_type"))) notamod(progname, module);
 if (!(slot[slotno].mod_name=(void*)GetProcAddress(slot[slotno].dll,"mod_name"))) notamod(progname, module);
 if (!(slot[slotno].mod_pref=(void*)GetProcAddress(slot[slotno].dll,"mod_pref"))) notamod(progname, module);
 if (!(slot[slotno].mod_vers=(void*)GetProcAddress(slot[slotno].dll,"mod_vers"))) notamod(progname, module);
 if (!(slot[slotno].mod_copr=(void*)GetProcAddress(slot[slotno].dll,"mod_copr"))) notamod(progname, module);
 
 printf ("Type: %c%c%c%c\n", ((char*)slot[slotno].mod_type)[0], ((char*)slot[slotno].mod_type)[1], 
                             ((char*)slot[slotno].mod_type)[2], ((char*)slot[slotno].mod_type)[3]);

 if ((*(slot[slotno].mod_type))!=0x544F4C53)
 {
  FreeLibrary(slot[slotno].dll);
  slot[slotno].dll=0;
  fprintf(stderr, "%s: %s is not a MoDapple slot module.\n", progname, module);
  return -1;
 }
 
 printf ("%s, Version %s\n%s\n", *(slot[slotno].mod_name), *(slot[slotno].mod_vers), *(slot[slotno].mod_copr));

 SFUNCTINO(slotno, preload_name, "preload_name");
 SFUNCTINO(slotno, preload_size, "preload_size");
 SFUNCTINO(slotno, preload_ptr,  "preload_ptr");

 SFUNCTINO(slotno, write16, "write16");
 SFUNCTINO(slotno, write256, "write256");
 SFUNCTINO(slotno, write2048, "write2048");

 SFUNCTINO(slotno, read16, "read16");
 SFUNCTINO(slotno, read256, "read256");
 SFUNCTINO(slotno, read2048, "read2048");

 SFUNCTINO(slotno, override_rd, "override_rd");
 SFUNCTINO(slotno, override_wr, "override_wr");
 
 SFUNCTINO(slotno, reset, "reset");
 SFUNCTINO(slotno, init, "init");
 SFUNCTINO(slotno, deinit, "deinit");

 SFUNCTINO(slotno, config, "config");
 
 if (*(slot[slotno].preload_ptr))
 {
  if (bload(*(slot[slotno].preload_name),*(slot[slotno].preload_size),*(slot[slotno].preload_ptr)))
  {
   FreeLibrary(slot[slotno].dll);
   slot[slotno].dll=0;
   return -1;
  }
 }
 
 e=slot[slotno].init();
 if (e==6502) /* plugin needs dma */
 {
  SFUNCTINO(slotno, dmainit, "dmainit");
  slot[slotno].dmainit(cpuwrite, cpuread);
 }
 else if (e)
 {
  fprintf (stderr, "%s: Could not initialize module %s\n", progname, module);
  FreeLibrary(slot[slotno].dll);
  slot[slotno].dll=0;
  return -1;
 }
 
 printf ("%s: Inserted in slot %d %s Version %s\n%s\n", progname,slotno,
         *(slot[slotno].mod_name),*(slot[slotno].mod_vers),*(slot[slotno].mod_copr));
 return 0;
}

int WINAPI WinMain(HINSTANCE ghinstance, HINSTANCE dummy, LPSTR cmdline, int nCmdShow)
{
 char inifile[1024];
 char modbuf[256], slotstr[6];
 char ttymod[256], cpumod[256];
 int s,l;
 int gotrom;
 
 hinstance=ghinstance;
 
 gotrom=0;
 
 getcwd(inifile,1010);
 strcat(inifile,"\\modapple.ini");
 
 for (l=0; l<8; l++) slot[l].dll=0;
 
 printf ("%s\n", verstr);
 
 l=*ttymod=*cpumod=0;
 
 l=GetPrivateProfileString("System", "Console", 0, modbuf, 256, inifile);
 if (l)
 {
  printf ("%s: Assigning Console to module %s\n", progname, modbuf);
  strcpy(ttymod, modbuf);
 }
 
 l=GetPrivateProfileString("System", "CPU", 0, modbuf, 256, inifile);
 if (l)
 {
  printf ("%s: Assigning CPU to module %s\n", progname, modbuf);
  strcpy(cpumod, modbuf);
 }
 
 for (s=0; s<8; s++)
 {
  sprintf (slotstr, "Slot%d", s);
  l=GetPrivateProfileString("SlotCards", slotstr, 0, modbuf, 256, inifile);
  if (l)
  {
   if (!slotassign(s,modbuf))
    printf ("%s: Linked %s to slot %d\n", progname, modbuf, s);
   else
    printf ("%s: Could not link %s to slot %d\n", progname, modbuf, s);
  }
 }
 
 if (!*ttymod)
 {
  fprintf (stderr, "%s: No console module provided\n", progname);
  return -1;
 }
 
 if (!*cpumod)
 {
  fprintf (stderr, "%s: No CPU module provided\n", progname);
  return -1;
 }
 
 mod_tty=0;
 mod_cpu=0;
 
 mod_tty=LoadLibrary(ttymod);
 if (!mod_tty)
 {
  fprintf(stderr, "%s: Could not open module %s\n", progname, ttymod);
  return -1;
 }
 printf ("%s: Opened module %s\n", progname, ttymod);
 if (!(tty_mod_type=(void*)GetProcAddress(mod_tty,"mod_type"))) notamod(progname, ttymod);
 if (!(tty_mod_name=(void*)GetProcAddress(mod_tty,"mod_name"))) notamod(progname, ttymod);
 if (!(tty_mod_pref=(void*)GetProcAddress(mod_tty,"mod_pref"))) notamod(progname, ttymod);
 if (!(tty_mod_vers=(void*)GetProcAddress(mod_tty,"mod_vers"))) notamod(progname, ttymod);
 if (!(tty_mod_copr=(void*)GetProcAddress(mod_tty,"mod_copr"))) notamod(progname, ttymod);
 
 printf ("Type: %c%c%c%c\n", ((char*)tty_mod_type)[0], ((char*)tty_mod_type)[1], 
                             ((char*)tty_mod_type)[2], ((char*)tty_mod_type)[3]);

 if ((*tty_mod_type)!=0x50595454)
 {
  FreeLibrary(mod_tty);
  fprintf(stderr, "%s: %s is not a MoDapple console module.\n", progname, ttymod);
  return -1;
 }
 
 printf ("%s, Version %s\n%s\n", *tty_mod_name, *tty_mod_vers, *tty_mod_copr);
 
 TFUNCTINO(ttykey,"ttykey");
 TFUNCTINO(ttywrite,"ttywrite");
 TFUNCTINO(ttyread,"ttyread");
 TFUNCTINO(ttyout,"ttyout");
 TFUNCTINO(ttyin,"ttyin");
 TFUNCTINO(ttydeinit,"ttydeinit");
 TFUNCTINO(ttyinit,"ttyinit");
 TFUNCTINO(ttyloop,"ttyloop");
 
 printf ("Link successful!\n");
 
 mod_cpu=LoadLibrary(cpumod);
 if (!mod_cpu)
 {
  fprintf(stderr, "%s: Could not open module %s\n", progname, cpumod);
  return -1;
 }
 printf ("%s: Opened module %s\n", progname, cpumod);
 if (!(cpu_mod_type=(void*)GetProcAddress(mod_cpu,"mod_type"))) notamod(progname, cpumod);
 if (!(cpu_mod_name=(void*)GetProcAddress(mod_cpu,"mod_name"))) notamod(progname, cpumod);
 if (!(cpu_mod_pref=(void*)GetProcAddress(mod_cpu,"mod_pref"))) notamod(progname, cpumod);
 if (!(cpu_mod_vers=(void*)GetProcAddress(mod_cpu,"mod_vers"))) notamod(progname, cpumod);
 if (!(cpu_mod_copr=(void*)GetProcAddress(mod_cpu,"mod_copr"))) notamod(progname, cpumod);
 
 printf ("Type: %c%c%c%c\n", ((char*)cpu_mod_type)[0], ((char*)cpu_mod_type)[1], 
                             ((char*)cpu_mod_type)[2], ((char*)cpu_mod_type)[3]);

 if ((*cpu_mod_type)!=0x32303536)
 {
  FreeLibrary(mod_tty);
  FreeLibrary(mod_cpu);
  fprintf(stderr, "%s: %s is not a MoDapple CPU module.\n", progname, cpumod);
  return -1;
 }
 
 printf ("%s, Version %s\n%s\n", *cpu_mod_name, *cpu_mod_vers, *cpu_mod_copr);

 CFUNCTINO(cpureset,"cpureset");
 CFUNCTINO(cpurun,"cpurun");
 CFUNCTINO(cpuinit,"cpuinit");
 CFUNCTINO(cpuseta,"cpuseta");
 CFUNCTINO(cpusetx,"cpusetx");
 CFUNCTINO(cpusety,"cpusety");
 CFUNCTINO(cpusets,"cpusets");
 CFUNCTINO(cpusetp,"cpusetp");
 CFUNCTINO(cpusetpc,"cpusetpc");
 CFUNCTINO(cpusetrate,"cpusetrate");
 CFUNCTINO(cpugeta,"cpugeta");
 CFUNCTINO(cpugetx,"cpugetx");
 CFUNCTINO(cpugety,"cpugety");
 CFUNCTINO(cpugets,"cpugets");
 CFUNCTINO(cpugetp,"cpugetp");
 CFUNCTINO(cpugetpc,"cpugetpc");
 CFUNCTINO(cpugetrate,"cpugetrate");

 printf ("Link successful!\n");
 
 l=GetPrivateProfileString("System", "BIOS", "apple2p.rom", modbuf, 256, inifile);

 if (!(bload(modbuf,12288,rom))) gotrom=1;
 
 if (!gotrom)
 {
  fprintf (stderr, "%s: No BIOS could be loaded.\n", progname);
 }
 
 holdc8=0;
 
 cpuinit(cpuwrite, cpuread, cpuloop);

 cpureset();
 cpusetrate(65);
 
 printf ("CPU start address: %04X\n", cpugetpc());
 
 *ttykey=nCmdShow;
 ttyinit();
 reset_all();
 cpurun();
 ttydeinit();
 for (l=7; l>=0; l--)
 {
  if (slot[l].dll) 
  {
   printf ("Releasing slot %d module.\n", l);
   slot[l].deinit();
  }
 }
 
 printf ("Releasing terminal.\n");
 FreeLibrary(mod_tty);
 printf ("Releasing CPU.\n");
 FreeLibrary(mod_cpu);
 
 printf ("Exiting cleanly.\n");
 return 0;
}
